@@ -1,86 +1,62 @@
-import numpy as np
-from matplotlib import pyplot as plt
-from scipy.fft import fft2, ifft2, rfft2, irfft2, set_global_backend
-from pprint import pprint
-import tqdm
-
-import pyfftw
 import time
 import threading
 import sys
-from typing import List
+import warnings
+from pprint import pprint
+import tqdm
+from typing import List, Optional
+
+import numpy as np
+from matplotlib import pyplot as plt
+from scipy.fft import fft2, ifft2, rfft2, irfft2, set_global_backend
+import pyfftw
 
 from michael960lib.math import fourier
-from michael960lib.common import overrides
-
-from torusgrid.fields import FreeEnergyFunctional2D, FieldMinimizer, RealField2D, NoiseGenerator2D, import_field, real_convolution_2d
+from michael960lib.common import overrides, experimental, deprecated
 from michael960lib.common import ModifyingReadOnlyObjectError, IllegalActionError
-
-from .base import PFCStateFunction, PFCFreeEnergyFunctional
+from torusgrid.fields import RealField2D, import_field
+from torusgrid.dynamics import FieldEvolver, FancyEvolver, NoiseGenerator2D, FreeEnergyFunctional2D, EvolverHistory
+from .base import PFCStateFunction, PFCFreeEnergyFunctional, import_state_function
 
 
-class PFCMinimizer(FieldMinimizer):
+
+
+class PFCMinimizer(FancyEvolver):
     def __init__(self, field: RealField2D, dt, eps):
         super().__init__(field)
         self.dt = dt
         self.eps = eps
         self.age = 0
-
-        self.label = 'NULL'
         self.history = PFCMinimizerHistory()
-        self.display_precision = 7
-    
-    def set_display_precision(self, display_precision):
-        self.display_precision = display_precision
 
-    @overrides(FieldMinimizer)
+        self.info['system'] = 'pfc'
+        self.info['minimizer'] = 'NULL'
+
+        self.info['dt'] = dt
+        self.info['eps'] = eps
+        self.info['label'] = self.label = f'NULL eps={eps} dt={dt}'
+
+        self.display_format = '[{label}] f={f:.5f} F={F:.5f} psibar={psibar:.5f}'
+    
+    @overrides(FancyEvolver)
+    def get_evolver_state(self):
+        return {'age': self.age}
+
+    @overrides(FancyEvolver)
     def on_create_progress_bar(self, progress_bar: tqdm.tqdm):
         progress_bar.set_description_str(f'[{self.label}]')
         
     def set_age(self, age):
         self.age = age
 
-    @overrides(FieldMinimizer)
-    def step(self):
-        raise NotImplementedError()
-
-    @overrides(FieldMinimizer)
-    def start(self):
-        super().start()
-        state_function = self.get_state_function()
-        self.history.append_state_function(self.age, state_function)
-
-    @overrides(FieldMinimizer)
-    def on_epoch_end(self, progress_bar: tqdm.tqdm):
-        dp = self.display_precision
-        state_function = self.get_state_function()
-        state_function_str = state_function.to_string(float_fmt=f'.{dp+4}f')
-        progress_bar.set_description_str(f'[{self.label}][t={self.age:.{dp}f}] {state_function_str}')
-        self.history.append_state_function(self.age, state_function)
-
-    @overrides(FieldMinimizer)
-    def on_nonstop_epoch_end(self):
-        dp = self.display_precision
-        state_function = self.get_state_function()
-        state_function_str = state_function.to_string(float_fmt=f'.{dp+4}f')
-        sys.stdout.write(f'\r[{self.label}][t={self.age:.{dp}f}] {state_function_str}')
-        self.history.append_state_function(self.age, state_function)
-
-    @overrides(FieldMinimizer)
-    def end(self):
-        super().end()
-        self.history.commit(self.label, self.field)
-
-    def get_state_function(self) -> PFCStateFunction:
-        raise NotImplementedError()
-
 
 class ConstantChemicalPotentialMinimizer(PFCMinimizer):
     def __init__(self, field: RealField2D, dt: float, eps: float, mu: float, noise_generator:NoiseGenerator2D=None):
         super().__init__(field, dt, eps)
-        self.label = f'const-mu eps={eps} mu={mu} dt={dt}'
+        self.info['minimizer'] = 'mu'
+        self.info['mu'] = self.mu = mu
+        self.info['label'] = self.label = f'mu eps={eps:.5f} mu={mu:.5f} dt={dt:.5f}'
    
-        self.mu = mu
         self.fef = PFCFreeEnergyFunctional(eps)
         self.noise_generator = noise_generator
 
@@ -93,12 +69,6 @@ class ConstantChemicalPotentialMinimizer(PFCMinimizer):
         self._exp_dt_kernel = np.exp(-dt*self._kernel)
         self._mu_dt_half = self.dt * self.mu / 2
 
-    @overrides(PFCMinimizer)
-    def set_display_precision(self, display_precision: int):
-        dp = display_precision
-        super().set_display_precision(dp)
-        self.label = f'const-mu eps={self.eps:.{dp+1}f} mu={self.mu:.{dp+1}f} dt={self.dt:.{dp}f}'
-
     @overrides(PFCMinimizer)
     def step(self):
         self.age += self.dt
@@ -130,7 +100,9 @@ class ConstantChemicalPotentialMinimizer(PFCMinimizer):
 class NonlocalConservedMinimizer(PFCMinimizer):
     def __init__(self, field: RealField2D, dt: float, eps: float, noise_generator:NoiseGenerator2D=None):
         super().__init__(field, dt, eps)
-        self.label = f'nonlocal eps={eps} dt={dt}'
+
+        self.info['label'] = self.label = f'nonlocal eps={eps:.5f} dt={dt:.5f}'
+        self.info['minimizer'] = 'nonlocal'
 
         self._kernel = 1-2*self.field.K2+self.field.K4 - self.eps
         self._exp_dt_kernel = np.exp(-dt*self._kernel)
@@ -143,43 +115,209 @@ class NonlocalConservedMinimizer(PFCMinimizer):
 
         self.psibar = np.mean(self.field.psi)
 
-    @overrides(PFCMinimizer)
-    def set_display_precision(self, display_precision: int):
-        dp = display_precision
-        super().set_display_precision(dp)
-        self.label = f'nonlocal eps={self.eps:.{dp+1}f} dt={self.dt:.{dp+1}f}'
-
-    #@overrides(PFCMinimizer)
-    def _step_deprecated(self):
-        self.age += self.dt
-        psi3k = rfft2(self.field.psi**3)
-
-        if self.is_noisy:
-            self.field.psi += self.dt * self.noise_generator.generate() 
-
-        self.field.fft2()
-        psik00 = self.field.psi_k[0,0]
-        self.field.psi_k -= self.dt * psi3k
-        self.field.psi_k *= self._exp_dt_kernel
-        self.field.psi_k[0,0] = psik00
-        self.field.ifft2()
-
     @overrides(PFCMinimizer)
     def step(self):
         self.age += self.dt
 
         if self.is_noisy:
             self.field.psi += self.dt * self.noise_generator.generate() 
-
+        
         self.field.psi /= np.sqrt(1+self.field.psi**2*self.dt)
+        self.field.psi += - np.mean(self.field.psi) + self.psibar
 
         self.field.fft2()
         self.field.psi_k *= self._exp_dt_kernel
         self.field.ifft2()
+        self.field.psi += - np.mean(self.field.psi) + self.psibar
 
         self.field.psi /= np.sqrt(1+self.field.psi**2*self.dt)
+        self.field.psi += - np.mean(self.field.psi) + self.psibar
+
+    @overrides(PFCMinimizer)
+    def get_state_function(self):
+        psibar = np.mean(self.field.psi)
+        f = self.fef.mean_free_energy_density(self.field)
+        F = self.fef.free_energy(self.field)
+        return PFCStateFunction(self.field.Lx, self.field.Ly, f, F, psibar)
+
+
+class NonlocalConservedRK4(NonlocalConservedMinimizer):
+    @experimental('Nonlocal RK4 is experimental and not optimized')
+    def __init__(self, field: RealField2D, dt: float, eps: float,
+            noise_generator: Optional[NoiseGenerator2D]=None,
+            k_regularizer=0.1, inertia=100):
+        super().__init__(field, dt, eps, noise_generator)
+        self.R = k_regularizer
+        self.info['minimizer'] = 'nonlocal-rk4'
+        self.info['M'] = inertia
+        self.info['R'] = k_regularizer
+        self.info['label'] = self.label = f'nonlocal-rk4 eps={eps} dt={dt} R={k_regularizer} M={inertia}'
+
+        self.field_tmp = self.field.copy()
+        self.field_tmp.initialize_fft()
+
+        self._deriv = RealField2D(field.Lx, field.Ly, field.Nx, field.Ny)
+        self._deriv.initialize_fft()
+
+        self.dfield = self.field.copy()
+        self.dfield.set_psi(0)
+        self.dfield.initialize_fft()
+
+        self.dfield_tmp = self.dfield.copy()
+        self.dfield_tmp.initialize_fft()
+
+        self.inertia = inertia
+
+    def _psi_dot(self):
+        self._deriv.psi[:,:] = -self.fef.derivative(self.field_tmp)
+        self._deriv.fft2()
+        self._deriv.psi_k *= np.exp(-self.R*self._deriv.K2)
+        self._deriv.ifft2()
+        F = self._deriv.psi - np.mean(self._deriv.psi)
+        return self.dfield_tmp.psi*self.inertia, -(self.dfield_tmp.psi - F)
+
+    # RK4
+    @overrides(NonlocalConservedMinimizer)
+    def step(self):
+        self.age += self.dt
+        self.field_tmp.psi[:] = self.field.psi
+        self.dfield_tmp.psi[:] = self.dfield.psi
+
+        k1, l1 = self._psi_dot()
+        self.field_tmp.psi += k1*self.dt/2
+        self.dfield_tmp.psi += l1*self.dt/2
+
+        k2, l2 = self._psi_dot()
+        self.field_tmp.psi += k2*self.dt/2
+        self.dfield_tmp.psi += l2*self.dt/2
+
+        k3, l3 = self._psi_dot()
+        self.field_tmp.psi += k3*self.dt
+        self.dfield_tmp.psi += l3*self.dt
+
+        k4, l4 = self._psi_dot()
+
+        self.field.psi += self.dt / 6 * (k1 + 2*k2 + 2*k3 + k4)
+        self.dfield.psi += self.dt / 6 * (l1 + 2*l2 + 2*l3 + l4)
+ 
+
+class NonlocalConservedRK4Plain(NonlocalConservedRK4):
+    def __init__(self, field, dt, eps,
+            noise_generator: Optional[NoiseGenerator2D]=None,
+            k_regularizer=0.1):
+        super().__init__(field, dt, eps, noise_generator, k_regularizer, 0)
+
+    def _psi_dot(self):
+        self._deriv.psi[:,:] = -self.fef.derivative(self.field_tmp)
+        self._deriv.fft2()
+        self._deriv.psi_k *= np.exp(-self.R*self._deriv.K2)
+        self._deriv.ifft2()
+        F = self._deriv.psi - np.mean(self._deriv.psi)
+        return F
+
+    @overrides(NonlocalConservedMinimizer)
+    def step(self):
+        self.age += self.dt
+        self.field_tmp.psi[:] = self.field.psi
+
+        k1 = self._psi_dot()
+        self.field_tmp.psi += k1*self.dt/2
+
+        k2 = self._psi_dot()
+        self.field_tmp.psi += k2*self.dt/2
+
+        k3 = self._psi_dot()
+        self.field_tmp.psi += k3*self.dt
+
+        k4 = self._psi_dot()
+
+        self.field.psi += self.dt / 6 * (k1 + 2*k2 + 2*k3 + k4)
+
+
+class NonlocalDescent(NonlocalConservedMinimizer):
+    def __init__(self, field: RealField2D, dt: float, eps: float,
+            noise_generator: Optional[NoiseGenerator2D]=None,
+            k_regularizer=0.1):
+        super().__init__(field, dt, eps, noise_generator)
+        self.R = k_regularizer
+        self.info['minimizer'] = 'nonlocal-desc'
+        self.info['R'] = k_regularizer
+        self.info['label'] = self.label = f'nonlocal-desc eps={eps} dt={dt} R={k_regularizer}'
+
+        self.field_tmp = self.field.copy()
+        self.field_tmp.initialize_fft()
+
+        self._deriv = RealField2D(field.Lx, field.Ly, field.Nx, field.Ny)
+        self._deriv.initialize_fft()
+
+    def _psi_dot(self):
+        self._deriv.psi[:,:] = -self.fef.derivative(self.field_tmp)
+        self._deriv.fft2()
+        self._deriv.psi_k *= np.exp(-self.R*self._deriv.K2)
+        self._deriv.ifft2()
+        F = self._deriv.psi - np.mean(self._deriv.psi)
+        return F
+
+    @overrides(NonlocalConservedMinimizer)
+    def step(self):
+        self.age += self.dt
+        self.field_tmp.psi[:,:] = self.field.psi[:,:]
+        F = self._psi_dot()
+        cutoff = 0.01 * self.dt
+
+        dT = self.dt / (np.sum(F**2)*self.field.dV + cutoff)
+        self.field.psi += F * dT
+
+
+class NonlocalTeleporter(NonlocalConservedRK4):
+    def __init__(self, dt, eps):
+        super().__init__(field, dt, eps, inertia=0)
+
+    @overrides(NonlocalConservedMinimizer)
+    def step(self):
+        raise NotImplementedError
+ 
+
+
+class ConservedMinimizer(PFCMinimizer):
+    @experimental('Local conserved minimizer is not implemented yet')
+    def __init__(self, field: RealField2D, dt: float, eps: float, noise_generator:NoiseGenerator2D=None):
+        super().__init__(field, dt, eps)
+        self.label = f'nonlocal eps={eps} dt={dt}'
+
+        self._kernel = 1-2*self.field.K2+self.field.K4 - self.eps
+        self._exp_dt_kernel = np.exp(-dt*self._kernel)
+        self.noise_generator = noise_generator
+        self.fef = PFCFreeEnergyFunctional(eps)
+        self.is_noisy = not (noise_generator is None)
+
+        if not field.fft_initialized():
+            field.initialize_fft()
+
+        self.psibar = np.mean(self.field.psi)
+
+        self.field_tmp = self.field.copy()
+
+    #@overrides(PFCMinimizer)
+    def step(self):
+        raise NotImplementedError
+        self.age += self.dt
+
+        if self.is_noisy:
+            self.field.psi += self.dt * self.noise_generator.generate() 
+        
+        self.field.psi /= np.sqrt(1+self.field.psi**2*self.dt)
+        self.field.psi += - np.mean(self.field.psi) + self.psibar
+
+        self.field.fft2()
+        self.field.psi_k *= self._exp_dt_kernel
+        self.field.ifft2()
+        self.field.psi += - np.mean(self.field.psi) + self.psibar
+
+        self.field.psi /= np.sqrt(1+self.field.psi**2*self.dt)
+        self.field.psi += - np.mean(self.field.psi) + self.psibar
+
 
-        self.field.psi -= np.mean(self.field.psi) - self.psibar
 
     @overrides(PFCMinimizer)
     def get_state_function(self):
@@ -195,17 +333,18 @@ class StressRelaxer(PFCMinimizer):
         super().__init__(field, dt, eps)
         self.f = 1
 
+        self.info['mu'] = mu
+        self.info['label'] = self.label = f'stress-relax eps={eps:.5f} mu={mu:.5f} dt={dt:.5f}'
+        self.info['minimizer'] = 'stress-relax'
+        self.info['expansion_rate'] = expansion_rate
 
         self.dt = dt
         self.eps = eps
         self.mu = mu
         self.noise_generator = noise_generator
         self.is_noisy = not (noise_generator is None)
-
         self.fef = PFCFreeEnergyFunctional(eps)
-
         self.dT = self.dt * expansion_rate / 2
-
         self.NN = field.Nx * field.Ny
 
         if not field.fft_initialized():
@@ -213,13 +352,6 @@ class StressRelaxer(PFCMinimizer):
 
         self._prepare_minimization()
 
-        self.set_display_precision(5)
-
-    @overrides(PFCMinimizer)
-    def set_display_precision(self, display_precision: int):
-        dp = display_precision
-        super().set_display_precision(dp)
-        self.label = f'stress-relax eps={self.eps:.{dp+1}f} mu={self.mu:.{dp+1}f} dt={self.dt:.{dp+1}f}'
 
     def _prepare_minimization(self):
         f = self.field
@@ -331,20 +463,76 @@ class StressRelaxer(PFCMinimizer):
         return r.sum(axis=(1,2)) * self._2_NN
 
 
-class PFCMinimizerHistory:
-    def __init__(self):
-        self.t = []
-        self.state_functions = []
-        self.age = 0
-        self.label = 'NULL'
-        self.final_field_state = None
-        self.committed = False 
+class RandomStepMinimizer(PFCMinimizer):
+    def __init__(self, field: RealField2D, dt: float, eps: float, mu: float, seed: int):
+        super().__init__(field, dt, eps)
+        self.target = 'Omega'
 
-    def append_state_function(self, t, sf: PFCStateFunction):
+        self.info['minimizer'] = 'random'
+        self.info['label'] = self.label = 'random eps={eps:.5f} mu={mu:.5f} dt={dt:.5f}'
+
+        if mu is None:
+            self.target = 'F'
+            self.label = 'random eps={eps} dt={dt}'
+
+        self.mu = mu
+        self.fef = PFCFreeEnergyFunctional(eps)
+
+        self.generator = np.random.default_rng(seed)
+        self.candidate_field = self.field.copy()
+
+        self.random_scale = self.dt**2 * np.sqrt(np.mean(self.fef.derivative(self.field)**2))
+        print(f'scale = {self.random_scale}')
+
+    @overrides(PFCMinimizer) 
+    def step(self):
+        self.age += self.dt
+        self.candidate_field.psi = self.field.psi + self.generator.normal(0, self.random_scale, size=(self.field.Nx,self.field.Ny))
+        if self.target == 'F':
+            F0 = self.fef.free_energy(self.field) 
+            F1 = self.fef.free_energy(self.candidate_field)
+            if F1 < F0:
+                self.field.psi[:,:] = self.candidate_field.psi
+        if self.target == 'Omega':
+            O0 = self.fef.grand_potential(self.field, self.mu) 
+            O1 = self.fef.grand_potential(self.candidate_field, self.mu)
+            if O1 < O0:
+                self.field.psi[:,:] = self.candidate_field.psi
+
+
+    @overrides(PFCMinimizer)
+    def get_state_function(self):
+        Lx = self.field.Lx
+        Ly = self.field.Ly
+        psibar = np.mean(self.field.psi)
+        f = self.fef.mean_free_energy_density(self.field)
+        F = self.fef.free_energy(self.field)
+
+        omega = None
+        Omega = None
+        if self.target == 'Omega':
+            omega = self.fef.mean_grand_potential_density(self.field, self.mu)
+            Omega = self.fef.grand_potential(self.field, self.mu)
+
+        return PFCStateFunction(Lx, Ly, f, F, psibar, omega, Omega)
+
+
+
+
+
+class PFCMinimizerHistory(EvolverHistory):
+    def __init__(self):
+        super().__init__()
+        self.t = []
+        self.age = 0
+
+    @overrides(EvolverHistory)
+    def append_state_function(self, evolver_state: dict, sf: PFCStateFunction):
         if self.committed:
             raise ModifyingReadOnlyObjectError(
-            f'history object (labe=\'{self.label}\') is already committed and hence not editable', self)
+            f'history object (label=\'{self.label}\') is already committed and hence not editable', self)
 
+        t = evolver_state['age']
         if t < self.age:
             raise IllegalActionError(f'time={t} is smaller than the current recorded time')
 
@@ -352,31 +540,9 @@ class PFCMinimizerHistory:
         self.state_functions.append(sf)
         self.age = t
 
-    def commit(self, label, field: RealField2D):
-        if self.committed:
-            raise IllegalActionError(f'history object (label=\'{self.label}\') is already committed')
-
-        self.label = label
-        self.committed = True 
-        self.final_field_state = field.export_state()
-
-    def is_committed(self):
-        return self.committed
-
     def get_t(self):
         return self.t
 
-    def get_state_functions(self) -> List[PFCStateFunction]:
-        return self.state_functions
-
-    def get_final_field_state(self):
-        if not self.committed:
-            raise IllegalActionError('cannot get final state from uncommitted PFC minimizer history')
-        return self.final_field_state
-
-    def get_label(self):
-        return self.label
-
     def export(self) -> dict:
         if self.committed:
             state = dict()
@@ -402,7 +568,4 @@ def import_minimizer_history(state: dict) -> PFCMinimizerHistory:
     return mh
 
 
- 
-
-
 
